package lambdapi

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module CommonSource

Sourcemodule Console : sig ... end

Flag management.

Sourcemodule Debug : sig ... end
Sourcemodule Error : sig ... end

Warnings and errors.

Sourcemodule Escape : sig ... end

Management of escaped identifiers "{|...|}".

Sourcemodule Library : sig ... end

Lambdapi library management.

Sourcemodule Logger : sig ... end
Sourcemodule Path : sig ... end

Module paths in the Lambdapi library.

Sourcemodule Pos : sig ... end

Source code position management. This module may be used to map sequences of characters in a source file to an abstract syntax tree.