package lambdapi

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Compile.PureSource

Pure wrappers around compilation functions. Functions provided perform the same computations as the ones defined earlier, but restores the state when they have finished. An optional library mapping or state can be passed as argument to change the settings.

Sourceval compile : ?lm:(Common.Path.t * string) -> ?st:Common.Console.State.t -> bool -> Common.Path.t -> Core.Sign.t
Sourceval compile_file : ?lm:(Common.Path.t * string) -> ?st:Common.Console.State.t -> string -> Core.Sign.t