package lambdapi

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Parsing.PrattSource

This module defines functions needed by the Pratt parser of the Pratter library.

The interface for the Pratter library can be seen at

Sourcemodule Pratt : sig ... end
include module type of struct include Pratt end

parse ss env t Pratt parses term t, unsugaring infix operators and prefix operators using signature state ss and environment env to determine which term is an operator, and to build new terms. Note that it doesn't recurse into abstractions or implications and alike.