package lambdapi

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Pratt.PrattSource

parse ss env t Pratt parses term t, unsugaring infix operators and prefix operators using signature state ss and environment env to determine which term is an operator, and to build new terms. Note that it doesn't recurse into abstractions or implications and alike.