package lambdapi

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module ToolSource

Sourcemodule External : sig ... end

Call to external checkers.

Sourcemodule Hrs : sig ... end

This module provides a function to translate a signature to the HRS format used in the confluence competition.

Sourcemodule Sr : sig ... end
Sourcemodule Tree_graphviz : sig ... end

Representation of trees as graphviz files.

Sourcemodule Xtc : sig ... end

This module provides a function to translate a signature to the XTC format used in the termination competition.